@@ -1,18 +1,33 @@
 """Test serializing ecdsa keys"""
 from __future__ import absolute_import, division, unicode_literals
+
 import unittest
 
+import pytest
+from helper import SkippableTest
+
 import jsonpickle
 
-from helper import SkippableTest
+
+@pytest.fixture(scope='module', autouse=True)
+def gmpy_extension():
+    """Initialize the gmpy extension for this test module"""
+    try:
+        jsonpickle.ext.gmpy.register_handlers()
+        yield  # control to the test function.
+        jsonpickle.ext.gmpy.unregister_handlers()
+    except AttributeError:
+        pytest.skip(
+            "gmpy was not detected, please try installing it for more complete tests!"
+        )
 
 
 class EcdsaTestCase(SkippableTest):
-
     def setUp(self):
         try:
             from ecdsa import NIST384p
             from ecdsa.keys import SigningKey
+
             self.NIST384p = NIST384p
             self.SigningKey = SigningKey
             self.should_skip = False
