@@ -5,40 +5,54 @@
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution.
 from __future__ import absolute_import, division, unicode_literals
+
+import collections
 import os
 import unittest
-import collections
+import warnings
 
 import pytest
+from helper import SkippableTest
 
 import jsonpickle
 import jsonpickle.backend
 import jsonpickle.handlers
 from jsonpickle import compat, tags, util
-from jsonpickle.compat import PY2, PY3
-from helper import SkippableTest
 
 
-class Thing(object):
+class ListLike:
+    def __init__(self):
+        self.internal_list = []
+
+    def append(self, item):
+        self.internal_list.append(item)
+
+    def __getitem__(self, i):
+        return self.internal_list[i]
+
+
+class Thing:
     def __init__(self, name):
         self.name = name
         self.child = None
 
+    def __iter__(self):
+        for attr in [
+            x for x in getattr(self.__class__, "__dict__") if not x.startswith("__")
+        ]:
+            yield attr, getattr(self, attr)
+
     def __repr__(self):
         return 'Thing("%s")' % self.name
 
 
-class Capture(object):
+class Capture:
     def __init__(self, *args, **kwargs):
         self.args = args
         self.kwargs = kwargs
 
 
-class ThingWithProps(object):
-    def __init__(self, name='', dogs='reliable', monkies='tricksy'):
-        self.name = name
-        self._critters = (('dogs', dogs), ('monkies', monkies))
-
+class ThingWithProps:
     def _get_identity(self):
         keys = [self.dogs, self.monkies, self.name]
         return hash('-'.join([str(key) for key in keys]))
@@ -55,6 +69,13 @@ class ThingWithProps(object):
 
     monkies = property(_get_monkies)
 
+    def __init__(self, name='', dogs='reliable', monkies='tricksy'):
+        self.name = name
+        self._critters = (('dogs', dogs), ('monkies', monkies))
+
+    def __eq__(self, other):
+        return self.identity == other.identity
+
     def __getstate__(self):
         out = dict(
             __identity__=self.identity,
@@ -74,9 +95,6 @@ class ThingWithProps(object):
         if ident != self.identity:
             raise ValueError('expanded object does not match original state!')
 
-    def __eq__(self, other):
-        return self.identity == other.identity
-
 
 class UserDict(dict):
     """A user class that inherits from :class:`dict`"""
@@ -86,12 +104,71 @@ class UserDict(dict):
         self.valid = False
 
 
-class Outer(object):
-    class Middle(object):
-        class Inner(object):
+class Outer:
+    class Middle:
+        class Inner:
             pass
 
 
+class MySlots:
+    __slots__ = ("alpha", "__beta")
+
+    def __init__(self):
+        self.alpha = 1
+        self.__beta = 1
+
+
+class MyPropertiesSlots:
+    __slots__ = ("alpha", "arr")
+
+    def __init__(self):
+        self.alpha = 1
+        self.arr = [1, 2, 3]
+
+    @property
+    def arr_len(self):
+        return len(self.arr)
+
+    @property
+    def other_object(self):
+        return Thing("Test")
+
+    def __eq__(self, other):
+        return (
+            self.alpha == other.alpha
+            and self.arr == other.arr
+            and self.arr_len == other.arr_len
+            and list(self.other_object) == list(other.other_object)
+        )
+
+
+class MyPropertiesDict:
+    def __init__(self):
+        self.alpha = 1
+        self.arr = [1, 2, 3]
+
+    @property
+    def arr_len(self):
+        return len(self.arr)
+
+    @property
+    def other_object(self):
+        return Thing("Test")
+
+    def __eq__(self, other):
+        return (
+            self.alpha == other.alpha
+            and self.arr == other.arr
+            and self.arr_len == other.arr_len
+            and list(self.other_object) == list(other.other_object)
+        )
+
+
+def on_missing_callback(class_name):
+    # not actually a runtime problem but it doesn't matter
+    warnings.warn("The unpickler couldn't find %s" % class_name, RuntimeWarning)
+
+
 class PicklingTestCase(unittest.TestCase):
     def setUp(self):
         self.pickler = jsonpickle.pickler.Pickler()
@@ -102,50 +179,30 @@ class PicklingTestCase(unittest.TestCase):
         self.pickler.reset()
         self.unpickler.reset()
 
-    @unittest.skipIf(not PY2, 'Python 2-specific base85 test')
-    def test_base85_always_false_on_py2(self):
-        self.assertFalse(self.b85_pickler.use_base85)
-
-    @unittest.skipIf(PY2, 'Base85 not supported on Python 2')
     def test_base85_override_py3(self):
         """Ensure the Python 2 check still lets us set use_base85 on Python 3"""
         self.assertTrue(self.b85_pickler.use_base85)
 
-    @unittest.skipIf(PY2, 'Base85 not supported on Python 2')
     def test_bytes_default_base85(self):
         data = os.urandom(16)
         encoded = util.b85encode(data)
         self.assertEqual({tags.B85: encoded}, self.b85_pickler.flatten(data))
 
-    @unittest.skipIf(PY2, 'Base85 not supported on Python 2')
     def test_py3_bytes_base64_default(self):
         data = os.urandom(16)
         encoded = util.b64encode(data)
         self.assertEqual({tags.B64: encoded}, self.pickler.flatten(data))
 
-    @unittest.skipIf(not PY2, 'Python 2-specific base64 test')
-    def test_py2_default_base64(self):
-        data = os.urandom(16)
-        encoded = util.b64encode(data)
-        self.assertEqual({tags.B64: encoded}, self.pickler.flatten(data))
-
-    @unittest.skipIf(PY2, 'Base85 not supported on Python 2')
     def test_decode_base85(self):
         pickled = {tags.B85: 'P{Y4;Xv4O{u^=-c'}
         expected = u'P\u00ffth\u00f6\u00f1 3!'.encode('utf-8')
         self.assertEqual(expected, self.unpickler.restore(pickled))
 
-    @unittest.skipIf(PY2, 'Base85 not supported on Python 2')
     def test_base85_still_handles_base64(self):
         pickled = {tags.B64: 'UMO/dGjDtsOxIDMh'}
         expected = u'P\u00ffth\u00f6\u00f1 3!'.encode('utf-8')
         self.assertEqual(expected, self.unpickler.restore(pickled))
 
-    @unittest.skipIf(not PY2, 'Python 2-specific base85 test')
-    def test_base85_crashes_py2(self):
-        with self.assertRaises(NotImplementedError):
-            self.unpickler.restore({tags.B85: 'P{Y4;Xv4O{u^=-c'})
-
     def test_string(self):
         self.assertEqual('a string', self.pickler.flatten('a string'))
         self.assertEqual('a string', self.unpickler.restore('a string'))
@@ -434,11 +491,85 @@ class PicklingTestCase(unittest.TestCase):
         no longer supported.
         """
         ae = jsonpickle.decode('{"py/type": "__builtin__.AssertionError"}')
-        assert ae is AssertionError
+        self.assertTrue(ae is AssertionError)
         ae = jsonpickle.decode('{"py/type": "exceptions.AssertionError"}')
-        assert ae is AssertionError
+        self.assertTrue(ae is AssertionError)
         cls = jsonpickle.decode('{"py/type": "__builtin__.int"}')
-        assert cls is int
+        self.assertTrue(cls is int)
+
+    def test_unpickler_on_missing(self):
+        class SimpleClass(object):
+            def __init__(self, i):
+                self.i = i
+
+        frozen = jsonpickle.encode(SimpleClass(4))
+        del SimpleClass
+
+        # https://docs.python.org/3/library/warnings.html#testing-warnings
+
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter("always")
+            jsonpickle.decode(frozen, on_missing='warn')
+            self.assertTrue(issubclass(w[-1].category, UserWarning))
+            self.assertTrue(
+                "Unpickler._restore_object could not find" in str(w[-1].message)
+            )
+
+            jsonpickle.decode(frozen, on_missing=on_missing_callback)
+            self.assertTrue(issubclass(w[-1].category, RuntimeWarning))
+            self.assertTrue("The unpickler couldn't find" in str(w[-1].message))
+
+        self.assertTrue(
+            jsonpickle.decode(frozen, on_missing='ignore')
+            == {
+                'py/object': 'jsonpickle_test.PicklingTestCase.test_unpickler_on_missing.<locals>.SimpleClass',
+                'i': 4,
+            }
+        )
+
+        try:
+            jsonpickle.decode(frozen, on_missing='error')
+        except jsonpickle.errors.ClassNotFoundError:
+            # it's supposed to error
+            self.assertTrue(True)
+        else:
+            self.assertTrue(False)
+
+    def test_private_slot_members(self):
+        obj = jsonpickle.loads(jsonpickle.dumps(MySlots()))
+        alpha = getattr(obj, "alpha", "(missing alpha)")
+        beta = getattr(obj, "_" + obj.__class__.__name__ + "__beta", "(missing beta)")
+        self.assertEqual(alpha, beta)
+
+    def test_include_properties_slots(self):
+        obj = MyPropertiesSlots()
+        dumped = jsonpickle.dumps(obj, include_properties=True)
+        self.assertTrue("py/property" in dumped)
+        self.assertEqual(obj, jsonpickle.loads(dumped))
+
+    def test_include_properties_dict(self):
+        obj = MyPropertiesDict()
+        dumped = jsonpickle.dumps(obj, include_properties=True)
+        self.assertTrue("py/property" in dumped)
+        self.assertEqual(obj, jsonpickle.loads(dumped))
+
+    def test_load_non_fully_qualified_classes(self):
+        # reuse MyPropertiesSlots because it has a nice eq method
+        obj = MyPropertiesSlots()
+        encoded = jsonpickle.encode(obj)
+        # MyPropertiesSlots and MyPropertiesDict have compatible eq methods
+        self.assertEqual(
+            obj,
+            jsonpickle.decode(encoded, classes={"MyPropertiesSlots": MyPropertiesDict}),
+        )
+
+    def test_warnings(self):
+        data = os.fdopen(os.pipe()[0])
+        with warnings.catch_warnings(record=True) as w:
+            jsonpickle.encode(data, warn=True)
+            assert len(w) == 1
+            assert issubclass(w[-1].category, UserWarning)
+            assert "replaced with None" in str(w[-1].message)
 
 
 class JSONPickleTestCase(SkippableTest):
@@ -495,10 +626,7 @@ class JSONPickleTestCase(SkippableTest):
         self.assertEqual(type(self.obj), type(actual))
 
     def test_unicode_dict_keys(self):
-        if PY2:
-            uni = unichr(0x1234)  # noqa
-        else:
-            uni = chr(0x1234)
+        uni = chr(0x1234)
         pickle = jsonpickle.encode({uni: uni})
         actual = jsonpickle.decode(pickle)
         self.assertTrue(uni in actual)
@@ -599,6 +727,21 @@ class JSONPickleTestCase(SkippableTest):
         self.assertEqual(obj.s, actual.s)
         self.assertEqual(obj.valid, actual.valid)
 
+    def test_dict_subclass_with_references(self):
+        d = UserDict()
+        d.s = 'string'
+        d['s'] = 'test'
+        obj = [d, d, d.__dict__]
+        pickle = jsonpickle.encode(obj)
+        actual = jsonpickle.decode(pickle)
+
+        self.assertEqual(type(actual), list)
+        self.assertEqual(type(actual[0]), UserDict)
+        self.assertEqual(actual[0].s, 'string')
+        self.assertEqual(actual[0]['s'], 'test')
+        self.assertIs(actual[0], actual[1])
+        self.assertIs(actual[0].__dict__, actual[2])
+
     def test_list_of_objects(self):
         """Test that objects in lists are referenced correctly"""
         a = Thing('a')
@@ -769,6 +912,10 @@ class JSONPickleTestCase(SkippableTest):
         assert decoded.a[1][0:3] == '[1,'
         assert decoded.a[2][0][0:3] == '[1,'
 
+    def _test_inner_class(self, InnerScope, obj, decoded):
+        self.assertTrue(isinstance(obj, InnerScope))
+        self.assertEqual(decoded.name, obj.name)
+
     def test_can_serialize_inner_classes(self):
         class InnerScope(object):
             """Private class visible to this method only"""
@@ -791,14 +938,7 @@ class JSONPickleTestCase(SkippableTest):
         decoded = jsonpickle.decode(encoded, classes={InnerScope})
         self._test_inner_class(InnerScope, obj, decoded)
 
-    def _test_inner_class(self, InnerScope, obj, decoded):
-        self.assertTrue(isinstance(obj, InnerScope))
-        self.assertEqual(decoded.name, obj.name)
-
     def test_can_serialize_nested_classes(self):
-        if PY2:
-            return self.skip('Serialization of nested classes requires ' 'Python >= 3')
-
         middle = Outer.Middle
         inner = Outer.Middle.Inner
         encoded_middle = jsonpickle.encode(middle)
@@ -813,9 +953,6 @@ class JSONPickleTestCase(SkippableTest):
         self.assertEqual(decoded_inner, inner)
 
     def test_can_serialize_nested_class_objects(self):
-        if PY2:
-            return self.skip('Serialization of nested classes requires ' 'Python >= 3')
-
         middle_obj = Outer.Middle()
         middle_obj.attribute = 5
         inner_obj = Outer.Middle.Inner()
@@ -831,10 +968,25 @@ class JSONPickleTestCase(SkippableTest):
         self.assertEqual(decoded_middle_obj.attribute, middle_obj.attribute)
         self.assertEqual(decoded_inner_obj.attribute, inner_obj.attribute)
 
+    def test_listlike(self):
+        """
+        https://github.com/jsonpickle/jsonpickle/issues/362
+        """
+        ll = ListLike()
+        ll.internal_list.append(1)
+        roundtrip_ll = jsonpickle.decode(jsonpickle.encode(ll))
+        self.assertEqual(len(ll.internal_list), len(roundtrip_ll.internal_list))
+
     def test_v1_decode(self):
         # TODO: Find a simple example that reproduces #364
         self.assertTrue(True)
 
+    def test_depth_tracking(self):
+        l = []
+        pickler = jsonpickle.Pickler()
+        pickler.flatten([l, l, l, l, l])
+        self.assertEqual(pickler._depth, -1)
+
 
 class PicklableNamedTuple(object):
     """
@@ -950,12 +1102,12 @@ class PickleProtocol2ReduceString(object):
 
 
 class PickleProtocol2ReduceExString(object):
-    def __reduce__(self):
-        assert False, "Should not be here"
-
     def __reduce_ex__(self, n):
         return __name__ + '.slotmagic'
 
+    def __reduce__(self):
+        assert False, "Should not be here"
+
 
 class PickleProtocol2ReduceTuple(object):
     def __init__(self, argval, optional=None):
@@ -974,12 +1126,12 @@ class PickleProtocol2ReduceTuple(object):
 
 @compat.iterator
 class ReducibleIterator(object):
-    def __iter__(self):
-        return self
-
     def __next__(self):
         raise StopIteration()
 
+    def __iter__(self):
+        return self
+
     def __reduce__(self):
         return ReducibleIterator, ()
 
@@ -1037,9 +1189,6 @@ class PickleProtocol2ReduceTupleState(PickleProtocol2ReduceTuple):
 
 
 class PickleProtocol2ReduceTupleSetState(PickleProtocol2ReduceTuple):
-    def __setstate__(self, state):
-        self.bar = state['foo']
-
     def __reduce__(self):
         return (
             type(self),  # callable
@@ -1049,6 +1198,9 @@ class PickleProtocol2ReduceTupleSetState(PickleProtocol2ReduceTuple):
             iter([]),  # dictitems
         )
 
+    def __setstate__(self, state):
+        self.bar = state['foo']
+
 
 class PickleProtocol2ReduceTupleStateSlots(object):
     __slots__ = ('argval', 'optional', 'foo')
@@ -1071,6 +1223,9 @@ class PickleProtocol2ReduceListitemsAppend(object):
     def __init__(self):
         self.inner = []
 
+    def append(self, item):
+        self.inner.append(item)
+
     def __reduce__(self):
         return (
             PickleProtocol2ReduceListitemsAppend,  # callable
@@ -1080,14 +1235,14 @@ class PickleProtocol2ReduceListitemsAppend(object):
             iter([]),  # dictitems
         )
 
-    def append(self, item):
-        self.inner.append(item)
-
 
 class PickleProtocol2ReduceListitemsExtend(object):
     def __init__(self):
         self.inner = []
 
+    def extend(self, items):
+        self.inner.exend(items)
+
     def __reduce__(self):
         return (
             PickleProtocol2ReduceListitemsAppend,  # callable
@@ -1097,14 +1252,14 @@ class PickleProtocol2ReduceListitemsExtend(object):
             iter([]),  # dictitems
         )
 
-    def extend(self, items):
-        self.inner.exend(items)
-
 
 class PickleProtocol2ReduceDictitems(object):
     def __init__(self):
         self.inner = {}
 
+    def __setitem__(self, k, v):
+        return self.inner.__setitem__(k, v)
+
     def __reduce__(self):
         return (
             PickleProtocol2ReduceDictitems,  # callable
@@ -1114,9 +1269,6 @@ class PickleProtocol2ReduceDictitems(object):
             iter(zip(['foo', 'bar'], ['foo', 'bar'])),  # dictitems
         )
 
-    def __setitem__(self, k, v):
-        return self.inner.__setitem__(k, v)
-
 
 class PickleProtocol2Classic:
     def __init__(self, foo):
@@ -1183,32 +1335,6 @@ class PicklingProtocol4TestCase(unittest.TestCase):
 
 
 class PicklingProtocol2TestCase(SkippableTest):
-    def test_classic_init_has_args(self):
-        """
-        Test unpickling a classic instance whose init takes args,
-        has no __getinitargs__
-        Because classic only exists under 2, skipped if PY3
-        """
-        if PY3:
-            return self.skip('No classic classes in PY3')
-        instance = PickleProtocol2Classic(3)
-        encoded = jsonpickle.encode(instance)
-        decoded = jsonpickle.decode(encoded)
-        self.assertEqual(decoded.foo, 3)
-
-    def test_getinitargs(self):
-        """
-        Test __getinitargs__ with classic instance
-
-        Because classic only exists under 2, skipped if PY3
-        """
-        if PY3:
-            return self.skip('No classic classes in PY3')
-        instance = PickleProtocol2ClassicInitargs(3)
-        encoded = jsonpickle.encode(instance)
-        decoded = jsonpickle.decode(encoded)
-        self.assertEqual(decoded.bar, 'choo')
-
     def test_reduce_complex_num(self):
         instance = 5j
         encoded = jsonpickle.encode(instance)
